from unittest import TestCase

from injector import Module, singleton

from zsl import inject
from zsl.application.containers.core_container import CoreContainer
from zsl.interface.task_queue import TaskQueueWorker, run_worker
from zsl.testing.db import IN_MEMORY_DB_SETTINGS
from zsl.testing.zsl import ZslTestCase, ZslTestConfiguration


class DummyWorker(TaskQueueWorker):

    def __init__(self):
        super(DummyWorker, self).__init__()
        self.run_count = 0

    def run(self):
        self.run_count += 1

    def stop_worker(self):
        pass


class DummyWorkerTaskQueueModule(Module):

    def configure(self, binder):
        worker = DummyWorker()
        binder.bind(TaskQueueWorker, to=worker, scope=singleton)


class DummyWorkerContainer(CoreContainer):
    dummy_worker = DummyWorkerTaskQueueModule


class RunDummyWorkerWithoutParamsTestCase(ZslTestCase, TestCase):
    CONFIG = IN_MEMORY_DB_SETTINGS.copy()
    ZSL_TEST_CONFIGURATION = ZslTestConfiguration(
        app_name='RunDummyWorkerWithoutParamsTestCase', container=DummyWorkerContainer,
        config_object=CONFIG)

    @inject(worker=TaskQueueWorker)
    def testWorkerRunCount(self, worker):
        self.assertEqual(worker.run_count, 0, "worker.run() shouldn't be called yet.")

        run_worker()

        self.assertEqual(worker.run_count, 1, "worker.run() should be called exactly once.")



"""
:mod:`zsl.db.helpers.pagination`
--------------------------------
"""
from __future__ import unicode_literals

from builtins import object
from typing import Dict, Union

from sqlalchemy.orm.query import Query

from zsl.db.model.app_model import AppModel

FIRST_PAGE = 1
DEFAULT_PAGE_SIZE = 25


class Pagination(object):
    """
    Pagination support. Allows to paginate a query. There are two choices.
      #. :meth:`.paginate` - paginates a query and obtains the count
         automatically.
      #. :meth:`.apply_pagination` - paginates a query and assumes that
         record count is set in advance.
    """

    def __init__(self, pagination=None):
        # type: (Union[PaginationRequest, Dict[str, Union[str, int]], None])->None
        pagination = self._create_pagination_request(pagination)
        assert isinstance(pagination, PaginationRequest)
        self._offset = (pagination.page_no - FIRST_PAGE) * pagination.page_size
        self._page_size = pagination.page_size
        self._record_count = -1

    def _create_pagination_request(self, pagination):
        # type: (Union[PaginationRequest, Dict[str, Union[str, int]], None])->PaginationRequest
        if pagination is None:
            pagination = PaginationRequest()
        elif isinstance(pagination, dict):
            page_size = int(pagination.get('page_size', DEFAULT_PAGE_SIZE))
            pagination = PaginationRequest(
                FIRST_PAGE + int(pagination.get('offset', 0)) // page_size,
                page_size)
        return pagination

    @property
    def record_count(self):
        return self._record_count

    @record_count.setter
    def record_count(self, record_count):
        self._record_count = record_count

    @property
    def page_size(self):
        return self._page_size

    @page_size.setter
    def page_size(self, page_size):
        # type: (int)->None
        self._page_size = page_size

    @property
    def offset(self):
        count = self._record_count
        per_page = self.page_size
        if self._offset >= count:
            last_page_size = count % per_page
            if last_page_size == 0:
                last_page_size = per_page
            self._offset = count - last_page_size
        if self._offset < 0:
            self._offset = 0

        return self._offset

    @offset.setter
    def offset(self, offset):
        # type:(int)->None
        self._offset = offset

    def apply_pagination(self, q):
        """
        Filters the query so that a given page is returned. The record count
        must be set in advance.
        :param q: Query to be paged.
        :return: Paged query.
        """
        # type: (Query)->Query
        assert self.record_count >= 0, "Record count must be set."
        return q.limit(self.page_size).offset(self.offset)

    def paginate(self, q):
        """
        Filters the query so that a given page is returned. The record count
        is computed automatically from query.
        :param q: Query to be paged.
        :return: Paged query.
        """
        self.record_count = q.count()
        return self.apply_pagination(q).all()


class PaginationRequest(AppModel):
    def __init__(self, page_no=FIRST_PAGE,
                 page_size=DEFAULT_PAGE_SIZE):
        super(PaginationRequest, self).__init__({})
        self.page_no = int(page_no) if page_no else FIRST_PAGE
        self.page_size = int(page_size) if page_size else DEFAULT_PAGE_SIZE


class PaginationResponse(AppModel):
    def __init__(self, record_count, page_size, pagination):
        # type: (int, int, PaginationRequest)->None
        super(PaginationResponse, self).__init__({})
        self.record_count = record_count
        max_page_size = pagination.page_size
        self.page_count = (record_count + max_page_size - 1) // max_page_size
        self.page_size = page_size
        self.max_page_size = max_page_size
        self.page_no = pagination.page_no



from __future__ import absolute_import, division, print_function, unicode_literals

from builtins import *
from collections import namedtuple

AlembicConfiguration = namedtuple('AlembicConfiguration', ['alembic_directory'])



from __future__ import absolute_import, division, print_function, unicode_literals

import json
from unittest.case import TestCase

from click.testing import CliRunner

from zsl import inject
from zsl.application.containers.core_container import CoreContainer
from zsl.application.modules.cli_module import ZslCli
from zsl.router.task import TaskConfiguration
from zsl.testing.db import IN_MEMORY_DB_SETTINGS
from zsl.testing.zsl import ZslTestCase, ZslTestConfiguration


class TestCliContainer(CoreContainer):
    pass


CONFIG = IN_MEMORY_DB_SETTINGS.copy()
CONFIG.update(
    TASKS=TaskConfiguration()
    .create_namespace("task")
    .add_packages(["zsl.tasks"])
    .get_configuration()
)


class ExecTaskFromCliTestCase(ZslTestCase, TestCase):
    ZSL_TEST_CONFIGURATION = ZslTestConfiguration(
        app_name="ExecTaskFromCliTestCase",
        config_object=CONFIG,
        container=TestCliContainer,
    )

    @inject(zsl_cli=ZslCli)
    def testRunningTestTask(self, zsl_cli):
        # type:(ZslCli)->None
        runner = CliRunner()
        result = runner.invoke(zsl_cli.cli, ["task", "task/zsl/test_task"])
        self.assertEqual(0, result.exit_code, "No error is expected.")
        self.assertEqual("ok", result.output.strip(), "Valid task output must be shown")

    @inject(zsl_cli=ZslCli)
    def testRunningTaskWithListInput(self, zsl_cli):
        # type:(ZslCli)->None
        runner = CliRunner()
        result = runner.invoke(
            zsl_cli.cli,
            ["task", "task/zsl/with_request_task", '{"list_of_numbers": [1,2,3] }'],
        )
        self.assertEqual(0, result.exit_code, "No error is expected.")
        self.assertEqual(
            json.loads("[1, 2, 3]"),
            json.loads(result.output.strip()),
            "Valid task output must be shown",
        )



import http.client
from unittest import TestCase

from zsl import Zsl, inject
from zsl.application.containers.web_container import WebContainer
from zsl.application.modules.web.cors import CORSConfiguration
from zsl.interface.web.performers.task import perform_web_task
from zsl.router.task import TaskConfiguration
from zsl.task.task_decorator import forbid_web_access
from zsl.testing.db import IN_MEMORY_DB_SETTINGS
from zsl.testing.zsl import ZslTestCase, ZslTestConfiguration


class HttpForbiddenTask(object):

    @forbid_web_access
    def perform(self, data):
        return "ok"


class ForbidWebAccessTestCase(ZslTestCase, TestCase):
    CONFIG = IN_MEMORY_DB_SETTINGS.copy()
    CONFIG.update(
        CORS=CORSConfiguration('origin'),
        TASKS=TaskConfiguration().create_namespace('n')
            .add_routes({'r': HttpForbiddenTask}).get_configuration())
    ZSL_TEST_CONFIGURATION = ZslTestConfiguration(
        app_name='ForbidWebAccessTestCase', container=WebContainer,
        config_object=CONFIG)

    @inject(app=Zsl)
    def testHttpAccessToWebForbiddenTask(self, app):
        """

        :param app: Zsl
        :return:
        """

        with app.test_request_context('/'):
            response = perform_web_task('n', 'r')
            self.assertEqual(http.client.FORBIDDEN,
                             response.status_code,
                             "Status code must be 403 (FORBIDDEN).")



"""
:mod:`zsl.interface.webservice.performers.default`
--------------------------------------------------

.. moduleauthor:: Martin Babka <babka@atteq.com>
"""
from __future__ import unicode_literals

from flask.globals import request

from zsl import Zsl, inject
from zsl.application.error_handler import error_handler
from zsl.interface.web.utils.execution import convert_to_web_response, notify_responders
from zsl.interface.web.utils.response_headers import append_headers
from zsl.router.task import RoutingError
from zsl.task.job_context import WebJobContext


@inject(app=Zsl)
def create_not_found_mapping(app):
    @app.route("/", defaults={'path': ''})
    @app.route("/<path:path>")
    @append_headers
    @notify_responders
    @convert_to_web_response
    @error_handler
    def not_found_mapping(path):
        """
        Default web request handler. Only returns 404 status code.
        """
        WebJobContext(path, None, None, None, request)
        raise RoutingError(path)



from typing import List, Union

CORS_CONFIGURATION_NAME = 'CORS'


class CORSConfiguration(object):
    DEFAULT_ALLOW_HEADERS = ['Accept', 'Origin', 'Content-Type',
                             'Authorization']
    DEFAULT_EXPOSE_HEADERS = ['Location', 'X-Total-Count', 'Link']
    DEFAULT_MAX_AGE = 21600

    def __init__(self, origin='', allow_headers=None, expose_headers=None,
                 max_age=DEFAULT_MAX_AGE):
        # type: (str, Union[List[str], str], Union[List[str], str], int)->None

        if allow_headers is None:
            allow_headers = self.DEFAULT_ALLOW_HEADERS

        if expose_headers is None:
            expose_headers = self.DEFAULT_EXPOSE_HEADERS

        self._origin = origin
        self._allow_headers = allow_headers
        self._expose_headers = expose_headers
        self._max_age = max_age

    @property
    def origin(self):
        # type: ()->str
        return self._origin

    @property
    def allow_headers(self):
        # type: ()->Union[List[str], str]
        return self._allow_headers

    @property
    def expose_headers(self):
        # type: ()->Union[List[str], str]
        return self._expose_headers

    @property
    def max_age(self):
        # type: ()->int
        return self._max_age



"""
:mod:`zsl.tasks.asl.version_task`
---------------------------------

Created on 18.02.2019

..moduleauthor:: Julius Flimmel
"""
from __future__ import unicode_literals

from builtins import object

from zsl import Zsl, inject
from zsl.interface.task import payload_into_model
from zsl.task.task_decorator import json_input, json_output


class WithRequestTask(object):

    class Request:
        def __init__(self):
            self.list_of_numbers = []  # type : List[int]

    """
    Returns received data if any or 'empty'
    """

    @inject(app=Zsl)
    def __init__(self, app):
        self._app = app

    @json_input
    @json_output
    @payload_into_model(Request)
    def perform(self, request):
        return request.list_of_numbers if request.list_of_numbers else "empty"



from abc import ABCMeta, abstractmethod
from typing import List


class ZslError(Exception):
    """
    A base class intended for all exceptions thrown by ZSL or ZSL application.
    """
    pass


ERROR_CONFIG_NAME = 'ERRORS'


class ErrorHandler(object):
    """
    Custom error handler providing a response on a particular error.
    """
    __metaclass__ = ABCMeta

    @abstractmethod
    def can_handle(self, e):
        """
        Indicator if the handler is able to handle the given exception `e`.

        :param e: The exception that shall be determined if can be handled by the handler.
        :return: `True` or `False` depending on whether the handler can/should handle the method.
        """
        # type: (ErrorHandler, Exception)->bool
        pass

    @abstractmethod
    def handle(self, e):
        """
        Handle the exception.

        :param e: The handled exception.
        :return: The error response for the exception.
        """
        # type: (ErrorHandler, Exception)->bool
        pass


class ErrorProcessor(object):
    """
    Custom error processor handling an error state caused by an error. For example an error processor may only log an
    error, send an email, etc. The main difference between an `ErrorProcessor` and an `ErrorHandler` is that the latter
    one returns a response. There may be only a single `ErrorHandler` returning whereas there may be more
    `ErrorProcessors` handling the same error.
    """
    __metaclass__ = ABCMeta

    @abstractmethod
    def handle(self, e):
        # type: (ErrorHandler, Exception)->None
        pass


class ErrorConfiguration(object):
    """
    The configuration object for error handling.
    """

    def __init__(self, handlers=None, processors=None, use_flask_handler=False):
        # type: (List[ErrorHandler], List[ErrorProcessor], bool)->None
        self._use_flask_handler = use_flask_handler
        self._handlers = handlers if handlers is not None else []
        self._processors = processors if processors is not None else []

    @property
    def use_flask_handler(self):
        # type:()->bool
        """
        In case of web requests, flasks provides a convenient way of exception
        handling. This handler shows the stack trace, etc. On the other hand \
        this setting will turn of ZSL's exception handling for web request.

        :return: Status
        """
        return self._use_flask_handler

    @property
    def handlers(self):
        # type: ()->List[ErrorHandler]
        return self._handlers

    @property
    def processors(self):
        # type: ()->List[ErrorProcessor]
        return self._processors



"""
:mod:`zsl.task.task_decorator`
------------------------------

.. moduleauthor:: Martin Babka <babka@atteq.com>,
                  Peter Morihladko <morihladko@atteq.com>,
                  Jan Janco <janco@atteq.com>,
                  Lubos Pis <pis@atteq.com>
"""
from __future__ import unicode_literals

from builtins import *
from datetime import timedelta
from functools import wraps
import json
import logging
from os.path import os
import traceback
from typing import Callable, List, Union

from flask import request

from zsl import Config, Injected, Zsl, inject
from zsl.application.modules.web.cors import CORSConfiguration
from zsl.constants import MimeType
from zsl.db.model import AppModelJSONEncoder
from zsl.task.job_context import JobContext, Responder, WebJobContext, add_responder
from zsl.task.task_data import TaskData
from zsl.utils.file_helper import makedirs
from zsl.utils.security_helper import verify_security_data
from zsl.utils.string_helper import join_list


def log_output(f):
    """
    Logs the output value.
    """

    @wraps(f)
    def wrapper_fn(*args, **kwargs):
        res = f(*args, **kwargs)
        logging.debug("Logging result %s.", res)
        return res

    return wrapper_fn


def save_to_file(destination_filename, append=False):
    """
    Save the output value to file.
    """

    def decorator_fn(f):
        @wraps(f)
        def wrapper_fn(*args, **kwargs):
            res = f(*args, **kwargs)

            makedirs(os.path.dirname(destination_filename))
            mode = "a" if append else "w"
            with open(destination_filename, mode) as text_file:
                text_file.write(res)

            return res

        return wrapper_fn

    return decorator_fn


def json_input(f):
    """
    Expects task input data in json format and parse this data.
    """

    @wraps(f)
    def json_input_decorator(*args, **kwargs):
        # If the data is already transformed, we do not transform it any
        # further.
        task_data = _get_data_from_args(args)

        if task_data is None:
            logging.error("Task data is empty during JSON decoding.")

        if task_data.payload:
            try:
                is_transformed = request.get_json()

                # We transform the data only in the case of plain POST requests.
                if not is_transformed:
                    task_data.transform_payload(json.loads)
            except (ValueError, RuntimeError):
                logging.error(
                    "Exception while processing JSON input decorator.")
                task_data.transform_payload(json.loads)
        else:
            task_data.transform_payload(lambda _: {})

        return f(*args, **kwargs)

    return json_input_decorator


def json_output(f):
    """
    Format response to json and in case of web-request set response content type
    to 'application/json'.
    """

    @wraps(f)
    def json_output_decorator(*args, **kwargs):
        @inject(config=Config)
        def get_config(config):
            return config

        config = get_config()

        rv = f(*args, **kwargs)
        indent = None
        if config.get('DEBUG', False):
            logging.getLogger(__name__).debug("Formatting JSON nicely.")
            indent = 2
        rv = json.dumps(rv, cls=AppModelJSONEncoder, indent=indent)
        _set_json_response_content_type()
        return rv

    return json_output_decorator


def _set_json_response_content_type():
    responder = MimeSetterWebTaskResponder(MimeType.APPLICATION_JSON.value)
    add_responder(responder)


def jsonp_wrap(callback_key='callback'):
    """
    Format response to jsonp and add a callback to JSON data - a jsonp request
    """

    def decorator_fn(f):

        @wraps(f)
        def jsonp_output_decorator(*args, **kwargs):
            task_data = _get_data_from_args(args)
            data = task_data.payload

            if callback_key not in data:
                raise KeyError(
                    'Missing required parameter "{0}" for task.'.format(
                        callback_key))

            callback = data[callback_key]
            jsonp = f(*args, **kwargs)
            if isinstance(JobContext.get_current_context(), WebJobContext):
                JobContext.get_current_context().add_responder(
                    MimeSetterWebTaskResponder('application/javascript'))
            jsonp = "{callback}({data})".format(callback=callback, data=jsonp)

            return jsonp

        return jsonp_output_decorator

    return decorator_fn


def jsend_output(fail_exception_classes=None):
    """
    Format task result to json output in jsend specification format. See:
    https://github.com/omniti-labs. Task return value must be dict or None.

    @param fail_exception_classes: exceptions which will produce 'fail' response
    status.
    """

    fail_exception_classes = fail_exception_classes if fail_exception_classes \
        else ()

    def decorator_fn(f):

        @wraps(f)
        @json_output
        def jsend_output_decorator(*args, **kwargs):
            try:
                rv = f(*args, **kwargs)
            except fail_exception_classes as e:
                return {'status': 'fail', 'data': {'message': str(e)}}
            except Exception as e:
                logging.error(str(e) + "\n" + traceback.format_exc())
                return {'status': 'error', 'message': 'Server error.'}

            if not isinstance(rv, dict) and rv is not None:
                msg = 'jsend_output decorator error: task must return dict ' \
                      'or None.\nTask return value: {0}.'
                logging.error(msg.format(rv))
                return {'status': 'error', 'message': 'Server error.'}

            return {'status': 'success', 'data': rv}

        return jsend_output_decorator

    return decorator_fn


def web_error_and_result(f):
    """
    Same as error_and_result decorator, except:
    If no exception was raised during task execution, ONLY IN CASE OF WEB
    REQUEST formats task result into json dictionary {'data': task return value}
    """

    @wraps(f)
    def web_error_and_result_decorator(*args, **kwargs):
        return error_and_result_decorator_inner_fn(f, True, *args, **kwargs)

    return web_error_and_result_decorator


def error_and_result(f):
    """
    Format task result into json dictionary `{'data': task return value}` if no
    exception was raised during the task execution. If there was raised an
    exception during task execution, formats task result into dictionary
    `{'error': exception message with traceback}`.
    """

    @wraps(f)
    def error_and_result_decorator(*args, **kwargs):
        return error_and_result_decorator_inner_fn(f, False, *args, **kwargs)

    return error_and_result_decorator


def error_and_result_decorator_inner_fn(f, web_only, *args, **kwargs):
    try:
        ret_val = f(*args, **kwargs)
        if web_only and not isinstance(JobContext.get_current_context(),
                                       WebJobContext):
            rv = ret_val
        else:
            rv = {'data': ret_val}
    except:  # NOQA
        exc = traceback.format_exc()
        logging.error(exc)
        rv = {'error': "{0}".format(exc)}

    return json.dumps(rv)


def required_data(*data):
    """
    Task decorator which checks if the given variables (indices) are stored
    inside the task data.
    """

    def decorator_fn(f):

        @wraps(f)
        def required_data_decorator(*args, **kwargs):
            task_data = _get_data_from_args(args).payload
            for i in data:
                if i not in task_data:
                    raise KeyError(
                        'Missing required parameter "{0}" for task.'.format(i))

            return f(*args, **kwargs)

        return required_data_decorator

    return decorator_fn


def append_get_parameters(accept_only_web=True):
    # type: (bool) -> Callable
    """
    Task decorator which appends the GET data to the task data.

    :param accept_only_web: Parameter which limits using this task only
                            with web requests.
    """

    def wrapper(f):

        @wraps(f)
        def append_get_parameters_wrapper_fn(*args, **kwargs):
            jc = JobContext.get_current_context()

            if isinstance(jc, WebJobContext):
                # Update the data with GET parameters
                web_request = jc.get_web_request()
                task_data = _get_data_from_args(args)
                data = task_data.payload
                data.update(web_request.args.to_dict(flat=True))
            elif accept_only_web:
                # Raise exception on non web usage if necessary
                raise Exception("append_get_parameters decorator may be used "
                                "with GET requests only.")

            return f(*args, **kwargs)

        return append_get_parameters_wrapper_fn

    return wrapper


def web_task(f):
    """
    Checks if the task is called through the web interface.
    Task return value should be in format {'data': ...}.
    """

    @wraps(f)
    def web_task_decorator(*args, **kwargs):
        jc = JobContext.get_current_context()
        if not isinstance(jc, WebJobContext):
            raise Exception(
                "The WebTask is not called through the web interface.")
        data = f(*args, **kwargs)
        jc.add_responder(WebTaskResponder(data))
        return data['data'] if 'data' in data else ""

    return web_task_decorator


def secured_task(f):
    """
    Secured task decorator.
    """

    @wraps(f)
    def secured_task_decorator(*args, **kwargs):
        task_data = _get_data_from_args(args)
        assert isinstance(task_data, TaskData)
        if not verify_security_data(task_data.payload['security']):
            raise SecurityException(
                task_data.payload['security']['hashed_token'])

        task_data.transform_payload(lambda x: x.get('data', {}))
        return f(*args, **kwargs)

    return secured_task_decorator


def xml_output(f):
    """
    Set content-type for response to WEB-REQUEST to 'text/xml'
    """

    @wraps(f)
    def xml_output_inner_fn(*args, **kwargs):
        ret_val = f(*args, **kwargs)

        if isinstance(JobContext.get_current_context(), WebJobContext):
            JobContext.get_current_context().add_responder(
                MimeSetterWebTaskResponder('text/xml'))
        return ret_val

    return xml_output_inner_fn


def file_upload(f):
    """
    Return list of `werkzeug.datastructures.FileStorage` objects - files to be
    uploaded
    """

    @wraps(f)
    def file_upload_decorator(*args, **kwargs):
        # If the data is already transformed, we do not transform it any
        # further.
        task_data = _get_data_from_args(args)

        if task_data is None:
            logging.error("Task data is empty during FilesUploadDecorator.")

        task_data.transform_payload(lambda _: request.files.getlist('file'))
        return f(*args, **kwargs)

    return file_upload_decorator


def _get_data_from_args(args):
    task_data = None
    for d in args:
        if isinstance(d, TaskData):
            task_data = d

    return task_data


class SecurityException(Exception):
    def __init__(self, hashed_token):
        Exception.__init__(self,
                           "Invalid hashed token '{0}'.".format(hashed_token))
        self._hashed_token = hashed_token

    def get_hashed_token(self):
        return self._hashed_token


class WebTaskResponder(Responder):
    def __init__(self, data):
        self.data = data

    def respond(self, response):
        for k in self.data:
            if k == 'headers':
                for header_name in self.data[k]:
                    response.headers[header_name] = self.data[k][header_name]
            else:
                setattr(response, k, self.data[k])


class MimeSetterWebTaskResponder(Responder):
    def __init__(self, mime):
        self._mime = mime

    def respond(self, r):
        r.content_type = self._mime


class CrossdomainWebTaskResponder(Responder):
    """
    source: http://flask.pocoo.org/snippets/56/
    """

    @inject(app=Zsl, cors_config=CORSConfiguration)
    def __init__(self, origin=None, methods=None, allow_headers=None,
                 expose_headers=None, max_age=None, app=Injected,
                 cors_config=Injected):
        # type: (str, List[str], str, str, Union[int, timedelta], Zsl, CORSConfiguration)->None
        self._app = app

        methods = join_list(methods, transform=lambda x: x.upper())
        self.methods = methods

        if allow_headers is None:
            allow_headers = cors_config.allow_headers
        allow_headers = join_list(allow_headers)
        self.allow_headers = allow_headers

        if expose_headers is None:
            expose_headers = cors_config.expose_headers
        expose_headers = join_list(expose_headers)
        self.expose_headers = expose_headers

        if origin is None:
            origin = cors_config.origin
        self.origin = join_list(origin)

        if max_age is None:
            max_age = cors_config.max_age
        if isinstance(max_age, timedelta):
            max_age = max_age.total_seconds()
        self.max_age = max_age

    def get_methods(self):
        if self.methods is not None:
            return self.methods

        options_resp = self._app.make_default_options_response()
        return options_resp.headers['allow']

    def respond(self, response):
        headers = response.headers
        headers['Access-Control-Allow-Origin'] = self.origin
        headers['Access-Control-Allow-Methods'] = self.get_methods()
        headers['Access-Control-Max-Age'] = str(self.max_age)
        headers['Access-Control-Allow-Headers'] = self.allow_headers
        headers['Access-Control-Expose-Headers'] = self.expose_headers
        return response


def crossdomain(origin=None, methods=None, allow_headers=None,
                expose_headers=None, max_age=None):
    def decorator(f):
        @wraps(f)
        def crossdomain_inner_fn(*args, **kwargs):
            responder = CrossdomainWebTaskResponder(
                origin, methods, allow_headers, expose_headers, max_age
            )

            rv = f(*args, **kwargs)
            add_responder(responder)
            return rv

        return crossdomain_inner_fn

    return decorator


def forbid_web_access(f):
    """
    Forbids running task using http request.

    :param f: Callable
    :return Callable
    """

    @wraps(f)
    def wrapper_fn(*args, **kwargs):
        if isinstance(JobContext.get_current_context(), WebJobContext):
            raise ForbiddenError('Access forbidden from web.')

        return f(*args, **kwargs)

    return wrapper_fn


class ForbiddenError(Exception):
    pass



from __future__ import absolute_import, division, print_function, unicode_literals

from builtins import *
import logging
from typing import Any

import click
from injector import Binder, provides, singleton

from zsl import Config, Zsl, inject
from zsl.application.initialization_context import InitializationContext
from zsl.application.modules.cli_module import ZslCli
from zsl.application.modules.context_module import DefaultContextModule, default_initializers
from zsl.application.modules.web.cors import CORS_CONFIGURATION_NAME, CORSConfiguration
from zsl.interface.web.performers.task import create_task_mapping
from zsl.utils.injection_helper import simple_bind


class WebInitializer(object):
    """Initialize the web application."""

    @staticmethod
    def initialize():
        """
        Import in this form is necessary so that we avoid the unwanted behavior and immediate initialization of the
        application objects. This makes the initialization procedure run in the time when it is necessary and has every
        required resources.
        """
        from zsl.interface.web.performers.default import create_not_found_mapping
        from zsl.interface.web.performers.resource import create_resource_mapping

        create_not_found_mapping()
        create_resource_mapping()


#: Initializers used in unit web applications
web_initializers = default_initializers + (WebInitializer,)


class WebCli(object):
    @inject(zsl_cli=ZslCli)
    def __init__(self, zsl_cli):
        # type: (ZslCli) -> None

        @zsl_cli.cli.group(help='Web related tasks.')
        def web():
            pass

        @web.command(help="Run web server and serve the application")
        @click.option('--host', '-h', help="host to bind to", default='127.0.0.1')
        @click.option('--port', '-p', help="port to bind to", default=5000)
        @inject(web_handler=WebHandler)
        def run(web_handler, host, port):
            # type: (WebHandler, str, int)->None
            web_handler.run_web(host=host, port=port)

        self._web = web

    @property
    def web(self):
        return self._web


class WebHandler(object):
    @inject(flask=Zsl)
    def run_web(self, flask, host='127.0.0.1', port=5000, **options):
        # type: (Zsl, str, int, **Any)->None
        """Alias for Flask.run"""
        return flask.run(
            host=flask.config.get('FLASK_HOST', host),
            port=flask.config.get('FLASK_PORT', port),
            debug=flask.config.get('DEBUG', False),
            **options
        )


class WebContextModule(DefaultContextModule):
    """Adds web application context to current configuration."""

    def _create_context(self):
        logging.getLogger(__name__).debug("Creating web context.")
        return InitializationContext(initializers=web_initializers)

    @provides(interface=WebCli, scope=singleton)
    def provide_web_cli(self):
        return WebCli()

    @provides(interface=WebHandler, scope=singleton)
    def provide_web_handler(self):
        return WebHandler()

    @provides(interface=CORSConfiguration, scope=singleton)
    @inject(config=Config)
    def provide_cors_configuration(self, config):
        # type: (Config)->CORSConfiguration
        return config.get(CORS_CONFIGURATION_NAME, CORSConfiguration())

    def configure(self, binder):
        # type: (Binder) -> None
        super(WebContextModule, self).configure(binder)
        simple_bind(binder, WebCli, singleton)
        create_task_mapping()



from unittest.case import TestCase

from zsl.utils.string_helper import camelcase_to_underscore, join_list, underscore_to_camelcase


class InflectionTestCase(TestCase):
    def testCamelCaseToUnderscore(self):
        self.assertEqual(
            "camel_case_to_underscore",
            camelcase_to_underscore("camelCaseToUnderscore"),
            "CC to usc conversion"
        )
        self.assertEqual(
            "camel_case_to_underscore",
            camelcase_to_underscore("CamelCaseToUnderscore"),
            "CC to usc conversion"
        )

    def testUnderscoreToCamelCase(self):
        self.assertEqual(
            "camelCaseToUnderscore",
            underscore_to_camelcase("camel_case_to_underscore", False),
            "CC to usc conversion"
        )
        self.assertEqual(
            "CamelCaseToUnderscore",
            underscore_to_camelcase("camel_case_to_underscore"),
            "CC to usc conversion"
        )
        self.assertEqual(
            "CamelCaseToUnderscore",
            underscore_to_camelcase("camel_case_to_underscore", True),
            "CC to usc conversion"
        )


class JoinListTestCase(TestCase):
    def testJoin(self):
        self.assertEqual('A, B', join_list(['a', 'b'], transform=lambda x: x.upper()))

    def testIdentity(self):
        self.assertEqual('a, b', join_list('a, b'))

    def testNone(self):
        self.assertEqual(None, join_list(None))



"""
:mod:`zsl.contrib.modules.sentry_module`
----------------------------------------

Sentry module is responsible for handling errors and sending them to a sentry
backend. Just provide a DSN (url) of the backend and you are all set up.

To add the Sentry support just add `SentryModule` to your container.

.. code-block:: python

    class MyApplicationContainer(WebContainer):
        sentry = SentryModule

Configuration
~~~~~~~~~~~~~

To configure one needs to provide a valid `SentryConfiguration` usually in
`default_settings.py` or in your profiles.

.. code-block:: python

    SENTRY = SentryConfiguration(dsn="...")

The following options are available:

    1. `dsn`: DSN/URl of your Sentry project,
    2. `environment`: the environment (release defaults to the version of your application),
    3. `tags`: additional tags str to str dictionary,
    4. `register_sentry_logging_handler`: bool value indicating if logging handler should be added to logging,
    5. `sentry_handler_logging_level`: minimal log level of log messages sent to sentry logging handler.

To test the functionality use CLI

.. code-block:: bash

    python app.py sentry test

"""
from __future__ import absolute_import, division, print_function, unicode_literals

from builtins import *

from flask import logging
from injector import Binder, provides, singleton
from raven import Client, setup_logging
from raven.handlers.logging import SentryHandler
from raven.transport import ThreadedRequestsHTTPTransport

from zsl import Config, Module, Zsl, inject
from zsl.application.error_handler import register
from zsl.application.modules.cli_module import ZslCli
from zsl.contrib.sentry.sentry_config import SentryConfiguration
from zsl.errors import ErrorProcessor, ZslError
from zsl.utils.injection_helper import simple_bind

try:
    import raven
except ImportError:
    CommandLine = None
    logging.getLogger(__name__).exception(
        "Can not import sentry client raven. Please install it first `pip install zsl ["
        "sentry]`.")
    raise


class SentryCli(object):
    pass

    """Sentry CLI interface support."""

    @inject(zsl_cli=ZslCli)
    def __init__(self, zsl_cli):
        # type: (ZslCli) -> None
        logging.getLogger(__name__).debug("Creating Sentry CLI.")

        @zsl_cli.cli.group()
        def sentry():
            pass

        @sentry.command(help='Send a test error to the Sentry backend.')
        def test():
            print('Sending test Sentry message.')
            raise ZslError("Sentry test error from Zsl.")

        self._sentry = sentry

    @property
    def sentry(self):
        return self._sentry


class SentryErrorProcessor(ErrorProcessor):

    @inject(config=SentryConfiguration, zsl=Zsl)
    def __init__(self, config, zsl):
        # type: (SentryConfiguration, Zsl)->None
        self._client = self._create_client(config, zsl)

        if config.register_logging_handler:
            self._register_logging_handler(config)

    @staticmethod
    def _create_client(config, zsl):
        # type: (SentryConfiguration, Zsl)->Client
        return Client(
            config.dsn,
            transport=ThreadedRequestsHTTPTransport,
            environment=config.environment,
            release=zsl.get_version(),
            tags=config.tags
        )

    def _register_logging_handler(self, config):
        # type: (SentryConfiguration)->SentryHandler
        handler = SentryHandler(self._client)
        handler.setLevel(config.sentry_logging_handler_level)
        setup_logging(handler)
        return handler

    def handle(self, e):
        logging.getLogger(__name__).info('Sending error message for {0}.'.format(e))
        self._client.captureException()


class SentryModule(Module):
    """Adds Sentry support."""

    SENTRY_CONFIG_NAME = 'SENTRY'

    @provides(SentryConfiguration)
    @inject(config=Config)
    def provide_sentry_configuration(self, config):
        # type: (Config) -> SentryConfiguration
        return config.get(SentryModule.SENTRY_CONFIG_NAME)

    def configure(self, binder):
        # type: (Binder) -> None
        simple_bind(binder, SentryCli, singleton)
        register(SentryErrorProcessor())



"""
:mod:`zsl.router.task`
------------------------
"""
from __future__ import absolute_import, division, print_function, unicode_literals

from abc import ABCMeta
from builtins import *
import importlib
import logging
from typing import Any, Callable, Dict, List, Tuple

from zsl import Config, inject
from zsl.errors import ZslError
from zsl.utils.string_helper import underscore_to_camelcase
from zsl.utils.task_helper import get_callable, instantiate

TASK_CONFIGURATION_NAME = 'TASKS'


class TaskNamespace(object):
    def __init__(self, namespace, task_configuration):
        # type: (str, TaskConfiguration)->None
        self._task_packages = []
        self._routes = {}
        self._task_configuration = task_configuration
        self._namespace = namespace

    def add_packages(self, packages):
        """
        Adds an automatic resolution of urls into tasks.
        :param packages: The url will determine package/module and the class.
        :return: self
        """
        # type: (List[str])->TaskNamespace
        assert isinstance(packages, list), "Packages must be list of strings."
        self._task_packages += packages
        return self

    def get_packages(self):
        # type:()->List[str]
        return list(self._task_packages)

    def add_routes(self, routes):
        """Adds the detailed mapping of urls to tasks.

        :param routes: Mapping which defines how urls of the namespace are
                       mapped to tasks. Each url (string) is mapped to a
                       callable which creates the task instance.

        :return: self
        """
        # type: (Dict[str, Callable])->TaskNamespace
        self._routes.update(routes)
        return self

    def get_routes(self):
        # type: ()->Dict[str, Callable]
        return self._routes.copy()

    def get_configuration(self):
        # type: ()->TaskConfiguration
        return self._task_configuration

    @property
    def namespace(self):
        # type:()->str
        return self._namespace


class TaskConfiguration(object):
    def __init__(self):
        self._namespaces = []  # type: List[TaskNamespace]

    def create_namespace(self, namespace):
        # type:(str)->TaskNamespace
        namespace = TaskNamespace(namespace, self)
        self._namespaces.append(namespace)
        return namespace

    @property
    def namespaces(self):
        return list(self._namespaces)


class RoutingError(ZslError):
    def __init__(self, path):
        msg = "Can not find task at path '{0}'.".format(path)
        super(RoutingError, self).__init__(msg)
        self._path = path

    @property
    def path(self):
        return self._path


class RouterStrategy(object):
    __metaclass__ = ABCMeta

    def can_route(self, path):
        # type:(str)->bool
        pass

    def route(self, path):
        # type:(str)->Callable
        pass


class PathTaskRouterStrategy(RouterStrategy):
    def __init__(self, task_configuration):
        # type: (TaskConfiguration)->None

        self._routes = {}
        for namespace_configuration in task_configuration.namespaces:
            namespace_routes = {}
            for k in namespace_configuration.get_routes():
                namespace_routes[namespace_configuration.namespace + '/' + k] = namespace_configuration.get_routes()[k]
            self._routes.update(namespace_routes)

    def can_route(self, path):
        return path in self._routes

    def route(self, path):
        return self._routes[path]


class PackageTaskRouterStrategy(RouterStrategy):
    def __init__(self, task_configuration, debug):
        # type: (TaskConfiguration)->None
        self._namespaces = task_configuration.namespaces
        self._debug = debug

    def can_route(self, path):
        return True

    def _load_module(self, module_name):
        # Debug loading provides better output.
        if self._debug:
            full = []
            for p in module_name.split('.'):
                full.append(p)
                importlib.import_module('.'.join(full))
        return importlib.import_module(module_name)

    def is_task_reloading(self):
        return self._debug

    def route(self, path):
        # type:(str)->Callable

        # Finding the path in task packages.
        logger = logging.getLogger(__name__)
        module_ = None
        exceptions = []
        class_name = None

        for task_namespace in self._namespaces:
            if not path.startswith(task_namespace.namespace):
                continue

            # Split the path into arrays - package names in the tasks package.
            class_name, package_path = self._split_path(path, task_namespace)
            task_packages = task_namespace.get_packages()
            module_, exceptions = self._find_task_in_namespace(task_packages, package_path, class_name)
            if module_ is not None:
                break

        if module_ is None:
            exception = RoutingError(path)
            logger.warning(str(exception))
            for e in exceptions:
                logger.error("Reason", exc_info=e)
            raise exception

        if self.is_task_reloading():
            importlib.reload(module_)
        cls = getattr(module_, class_name)
        return cls

    def _split_path(self, path, task_namespace):
        # type:(str, TaskNamespace)->Tuple[str, List[str]]
        package_path = path[len(task_namespace.namespace + '/'):]
        package_path = package_path.split("/")
        class_name = underscore_to_camelcase(package_path[-1])
        return class_name, package_path

    def _find_task_in_namespace(self, task_packages, package_path, class_name):
        logger = logging.getLogger(__name__)
        exceptions = []
        module_ = None
        for task_package in task_packages:
            module_name = "{0}.{1}".format(task_package, ".".join(package_path))

            try:
                logger.debug("Trying to load module with name '%s' and class name '%s'.",
                             module_name,
                             class_name)
                module_ = self._load_module(module_name)
                break
            except ImportError as e:
                exceptions.append(e)
                if self._debug:
                    logger.warning(
                        "Could not load module with name '%s' and class name '%s', '%s'; proceeding to next module.",
                        module_name, class_name, e)
        return module_, exceptions


class TaskRouter(object):
    @inject(config=Config, task_configuration=TaskConfiguration)
    def __init__(self, config, task_configuration):
        # type: (Config, TaskConfiguration) -> None
        self._config = config
        self._task_configuration = task_configuration  # type: TaskConfiguration
        self._strategies = [
            PathTaskRouterStrategy(self._task_configuration),
            PackageTaskRouterStrategy(self._task_configuration, self._config.get('DEBUG', False))
        ]

    def route(self, path):
        # type: (str)->Tuple[Any, Callable]
        """
        Returns the task handling the given request path.
        """
        logging.getLogger(__name__).debug("Routing path '%s'.", path)

        cls = None
        for strategy in self._strategies:
            if strategy.can_route(path):
                cls = strategy.route(path)
                break

        if cls is None:
            raise RoutingError(path)

        return self._create_result(cls)

    def _create_result(self, cls):
        # type:(Callable)->Tuple[Any, Callable]
        """
        Create the task using the injector initialization.
        :param cls:
        :return:
        """
        task = instantiate(cls)
        logging.getLogger(__name__).debug("Task object {0} created [{1}].".format(cls.__name__, task))
        return task, get_callable(task)



from __future__ import absolute_import, division, print_function, unicode_literals

from builtins import *  # NOQA
import http.client
from unittest.case import TestCase

from zsl import Zsl, inject
from zsl.application.containers.web_container import WebContainer
from zsl.application.modules.web.cors import CORSConfiguration
from zsl.interface.web.performers.task import perform_web_task
from zsl.router.task import TaskConfiguration
from zsl.task.task_decorator import crossdomain
from zsl.testing.db import IN_MEMORY_DB_SETTINGS
from zsl.testing.zsl import ZslTestCase, ZslTestConfiguration


class TestTask(object):
    @crossdomain('custom-origin', 'm1', 'allow-h', 'expose-h', 21)
    def perform(self, _data):
        return "ok"


class CorsTestCase(ZslTestCase, TestCase):
    CONFIG = IN_MEMORY_DB_SETTINGS.copy()
    CONFIG.update(
        CORS=CORSConfiguration('default-origin', ['default-allow-headers'], ['default-expose-headers'], 42),
        TASKS=TaskConfiguration()
            .create_namespace('n')
                .add_routes({'r': TestTask})
                .get_configuration().
            create_namespace('z')
                .add_packages(['zsl.tasks.zsl'])
                .get_configuration()
    )
    ZSL_TEST_CONFIGURATION = ZslTestConfiguration(
        app_name='CorsTestCase', container=WebContainer,
        config_object=CONFIG)

    @inject(app=Zsl)
    def testNotFound(self, app):
        with app.test_request_context('/'):
            response = perform_web_task('nn', 'r')

            self.assertEqual(http.client.NOT_FOUND,
                             response.status_code,
                             "Status code must be NOT_FOUND.")

            self.assertEqual('default-allow-headers', response.headers['Access-Control-Allow-Headers'])
            self.assertEqual('default-expose-headers', response.headers['Access-Control-Expose-Headers'])
            self.assertEqual('default-origin', response.headers['Access-Control-Allow-Origin'])
            self.assertEqual('42', response.headers['Access-Control-Max-Age'])


    @inject(app=Zsl)
    def testCustom(self, app):
        with app.test_request_context('/'):
            response = perform_web_task('n', 'r')

            self.assertEqual(http.client.OK,
                             response.status_code,
                             "Status code must be OK.")

            self.assertEqual('allow-h', response.headers['Access-Control-Allow-Headers'])
            self.assertEqual('expose-h', response.headers['Access-Control-Expose-Headers'])
            self.assertEqual('custom-origin', response.headers['Access-Control-Allow-Origin'])
            self.assertEqual('21', response.headers['Access-Control-Max-Age'])



"""
This script is to specify an appropriate module working with Python 3 (python3_gearman) or Python 2 (gearman).
"""

try:
    import python3_gearman as gearman
except ImportError:
    import gearman



from __future__ import absolute_import, division, print_function, unicode_literals

from unittest import TestCase

from zsl.db.model import AppModel
from zsl.db.model.app_model import RELATED_FIELDS, RELATED_FIELDS_CLASS


class TreeNodeAppModel(AppModel):
    pass


class CreateSimpleModelTestCase(TestCase):

    def testAppModelHints(self):
        model = TreeNodeAppModel(
            {
                'left': {'left': 0, 'op': '+', 'right': 1},
                'op': '*',
                'right': {'left': 3, 'op': '-', 'right': 2}
            },
            hints={
                RELATED_FIELDS: {
                    'left': {
                        RELATED_FIELDS_CLASS: TreeNodeAppModel
                    },
                    'right': {
                        RELATED_FIELDS_CLASS: TreeNodeAppModel
                    }
                }
            }
        )

        self.assertEqual('*', model.op)
        self.assertIsNotNone(model.left)
        self.assertIsNotNone(model.right)

        self.assertEqual('+', model.left.op)
        self.assertEqual('-', model.right.op)

        self.assertEqual(0, model.left.left)
        self.assertEqual(1, model.left.right)

        self.assertEqual(3, model.right.left)
        self.assertEqual(2, model.right.right)



from typing import Any

_current_app = None


def get_current_app():
    # type: () -> Any
    return _current_app


def set_current_app(app):
    # type: (Any) -> None
    global _current_app

    _current_app = app



"""
:mod:`zsl.db.model.app_model`
-----------------------------

.. moduleauthor:: Martin Babka
"""
from __future__ import unicode_literals

# noinspection PyCompatibility
from builtins import map, object

from future.utils import viewitems

from zsl.utils.dict_to_object_conversion import extend_object_by_dict

DATE_DATA = 'date_data'
DATETIME_DATA = 'datetime_data'
RELATED_FIELDS = 'related_fields'
RELATED_FIELDS_CLASS = 'cls'
RELATED_FIELDS_HINTS = 'hints'

ISO_8601_FORMAT = '%Y-%m-%dT%H:%M:%S.%fZ'


class AppModel(object):
    """AppModel's are used as a thin and simple communication objects. Also
    they can be saved into cache. Basically they are known as Data Transfer
    Objects or DTOs.

    .. automethod:: __init__
    """

    _not_serialized_attributes = ['_not_serialized_attributes', '_hints',
                                  '_id_name']

    def __init__(self, raw, id_name='id', hints=None):
        """
        The application model model constructor.

        :param raw: Dictionary of properties of the raw data.
        :param id_name: Name of the identifier property.
        :param hints: Tells which of the raw attributes are date or datetime
                      string and what is theirs format. Example:
                      ```
                      {
                          DATE_DATA: { 'birthday': '%d.%m.%Y' },
                          DATETIME_DATA: { 'created': '%Y-%m-%d %H:%M:%S' }
                      }
                      ```
                      this attributes are then saved in the standard zsl
                      service date/datetime format (consult
                      :mod:`zsl.utils.date_helper` for more.)
        """

        extend_object_by_dict(self, raw, hints)
        self._id_name = id_name

    def get_id(self):
        return self.__dict__[self._id_name]

    def _set_id_name(self, id_name):
        self._id_name = id_name

    @staticmethod
    def convert(v):
        if isinstance(v, AppModel):
            return v.get_attributes()
        else:
            return v

    def get_attributes(self):
        d = dict(self.__dict__)

        for k in self.__dict__:
            if k in self._not_serialized_attributes:
                d.pop(k)

            elif isinstance(d[k], AppModel):
                d[k] = self.convert(d[k])

            elif isinstance(d[k], list):
                d[k] = list(map(self.convert, d[k]))

            elif isinstance(d[k], tuple):
                d[k] = list(map(self.convert, d[k]))
                d[k] = tuple(d[k])

            elif isinstance(d[k], dict):
                for key, value in viewitems(getattr(self, k)):
                    d[k][key] = self.convert(value)

        return d

    def __str__(self):
        return "{0}: {1}".format(self.__class__, self.__dict__)



from __future__ import unicode_literals

# noinspection PyCompatibility
from builtins import int, str
from datetime import date, datetime
from typing import Any, Dict, Union

from future.utils import viewitems

from zsl.utils.date_helper import format_date_portable, format_datetime_portable

DATE_DATA = 'date_data'
DATETIME_DATA = 'datetime_data'
RELATED_FIELDS = 'related_fields'
RELATED_FIELDS_CLASS = 'cls'
RELATED_FIELDS_HINTS = 'hints'


def extend_object_by_dict(target, dict_data, hints=None):
    hints = _get_hints(hints)

    for field_name, field_value in viewitems(dict_data):
        if isinstance(field_value, (type(None), str, int, float, bool)):
            if field_name in hints[DATE_DATA] and field_value is not None:
                d = datetime.strptime(field_value,
                                      hints[DATE_DATA][field_name]).date()
                setattr(target, field_name, format_date_portable(d))
            elif field_name in hints[DATETIME_DATA] and \
                    field_value is not None:
                d = datetime.strptime(field_value,
                                      hints[DATETIME_DATA][field_name])
                setattr(target, field_name, format_datetime_portable(d))
            else:
                setattr(target, field_name, field_value)
        elif isinstance(field_value, datetime):
            setattr(target, field_name, format_datetime_portable(field_value))
        elif isinstance(field_value, date):
            setattr(target, field_name, format_date_portable(field_value))
        elif field_name in hints[RELATED_FIELDS]:
            related_cls = hints[RELATED_FIELDS][field_name][
                RELATED_FIELDS_CLASS]
            related_hints = hints[RELATED_FIELDS][field_name].get(
                RELATED_FIELDS_HINTS)
            if isinstance(field_value, (list, tuple)):
                setattr(
                    target,
                    field_name,
                    [
                        related_cls(_to_dict(x), 'id', related_hints)
                        for x in field_value
                    ]
                )
            else:
                setattr(
                    target,
                    field_name,
                    related_cls(_to_dict(field_value), 'id', related_hints)
                )
        elif isinstance(field_value, (list, tuple)):
            setattr(target, field_name, [x for x in field_value])


def _to_dict(val):
    # type: (Union[Dict[str, Any], object]) -> Dict[str, Any]
    return val if isinstance(val, dict) else val.__dict__


def _get_hints(original_hints):
    if original_hints is None:
        correct_hints = {DATE_DATA: {}, DATETIME_DATA: {}, RELATED_FIELDS: {}}
    else:
        correct_hints = original_hints
        if DATE_DATA not in correct_hints:
            correct_hints[DATE_DATA] = {}
        if DATETIME_DATA not in correct_hints:
            correct_hints[DATETIME_DATA] = {}
        if RELATED_FIELDS not in correct_hints:
            correct_hints[RELATED_FIELDS] = {}

    return correct_hints



"""
:mod:`zsl.utils.resource_helper`
--------------------------------

Helper module for resource management.
"""
# TODO describe what model resource is and use cases

from __future__ import unicode_literals

import importlib
import logging
from typing import Callable, Dict, List, Union

from flask import request

from zsl import Config, Injected, Zsl, inject
from zsl.db.model import AppModel
from zsl.interface.resource import ResourceResult
from zsl.resource.model_resource import ModelResource
from zsl.utils.string_helper import underscore_to_camelcase
from zsl.utils.task_helper import instantiate


class MethodNotImplementedException(Exception):
    """Exception raised on missing method"""
    pass


def parse_resource_path(path):
    """Split the path to its elements.

    :param path: URL path
    :type path: str
    :return: name and rest of the path
    :rtype: tuple(str, list(str))
    """
    splits = path.split('/')

    return splits[0], splits[1:]


def get_method(resource, method):
    """Test and return a method by name on resource.

    :param resource: resource object
    :type resource: object
    :param method: method name
    :type method: str
    :return: bounded method
    :raises MethodNotImplementedException: when method not found
    """
    if hasattr(resource, method) and callable(getattr(resource, method)):
        return getattr(resource, method)
    else:
        raise MethodNotImplementedException()


@inject(config=Config)
def get_resource_task(resource_path, config=Injected):
    # type: (str, Config) -> Callable[[str, Dict, Dict], Union[List[AppModel], AppModel, ResourceResult]]
    """Search and return a bounded method for given path.

    :param resource_path: resource path
    :type resource_path: str
    :param config: current configuration, injected
    :type config: Config
    :return: bounded method or None when not found
    :raises NameError: when can't find given resource by path
    """
    class_name = underscore_to_camelcase(resource_path) + 'Resource'

    resource_packages = config['RESOURCE_PACKAGES'] if 'RESOURCE_PACKAGES' in config else (config['RESOURCE_PACKAGE'],)

    resource = None
    for resource_package in resource_packages:
        module_name = "{0}.{1}".format(resource_package, resource_path)

        try:
            module = importlib.import_module(module_name)

            try:
                cls = getattr(module, class_name)

                resource = instantiate(cls)
                break
            except AttributeError:
                raise NameError("Can't find resource [{0}]".format(resource_path))

        except ImportError:

            # try to find __exposer__
            try:
                exposer = importlib.import_module(resource_package + '.__exposer__')

                resource = exposer.get_resource(class_name)
                break
            except Exception as e:
                logging.error("Can not load resource {0} [{1}].".format(resource_path, e))
                pass

    if resource is None:
        raise NameError("Can't find resource [{0}]".format(resource_path))

    try:
        if request.method == 'POST':
            return get_method(resource, 'create')

        elif request.method == 'GET':
            return get_method(resource, 'read')

        elif request.method == 'PUT' or request.method == 'PATCH':
            return get_method(resource, 'update')

        elif request.method == 'DELETE':
            return get_method(resource, 'delete')

        else:
            logging.error("Invalid request method [%s] is requested for path [%s]", request.method, resource_path)

    except MethodNotImplementedException:
        logging.error(
            "MethodNotImplementedException raised for method [%s] and path [%s]",
            request.method,
            resource_path
        )
        return None


@inject(app=Zsl)
def create_model_resource(resource_map, name, app=Injected):
    """Create a model resource from a dict ``resource_map``
    {'resource name': ('model package', 'model class')}

    :param resource_map: dict with resource descriptions
    :type resource_map: dict(str, tuple(str))
    :param name: name of the concrete resource
    :param app: current application, injected
    :type app: Zsl
    :return: instantiated model resource
    """
    try:
        resource_description = resource_map[name]
        if len(resource_description) == 2:
            module_name, model_name = resource_map[name]
            resource_class = ModelResource
        elif len(resource_description) == 3:
            module_name, model_name, resource_class = resource_map[name]
        else:
            raise ImportError("Invalid resource description for resource '{0}'".format(name))
    except KeyError:
        raise ImportError("Missing resource description for resource '{0}'".format(name))

    module = importlib.import_module(module_name)
    model_cls = getattr(module, model_name)

    return app.injector.create_object(resource_class, {'model_cls': model_cls})



from injector import Binder, Module, provides, singleton

from zsl import Config, inject
from zsl.application.error_handler import register
from zsl.errors import ERROR_CONFIG_NAME, ErrorConfiguration


class ErrorHandlerModule(Module):
    @provides(interface=ErrorConfiguration, scope=singleton)
    @inject(config=Config)
    def provide_error_config(self, config):
        # type: (Config)->ErrorConfiguration
        return config.get(ERROR_CONFIG_NAME, ErrorConfiguration())

    def configure(self, binder):
        @inject(error_config=ErrorConfiguration)
        def get_error_config(error_config):
            # type: (ErrorConfiguration)->ErrorConfiguration
            return error_config
            # type: (Binder)->None

        super(ErrorHandlerModule, self).configure(binder)
        for handler in get_error_config().handlers:
            register(handler)



"""
:mod:`zsl.application.error_handler`
---------------------------------------------------

This module does the error handling. It allows users to register
an error handler for a given exception type. It also provides default
error handlers.
"""
from __future__ import absolute_import, division, print_function, unicode_literals

from builtins import *
from functools import wraps
import http.client
import logging
import traceback
from typing import List

from flask import request

from zsl import inject
from zsl.db.model.app_model import AppModel
from zsl.errors import ErrorConfiguration, ErrorHandler, ErrorProcessor
from zsl.interface.task import ModelConversionError
from zsl.router.task import RoutingError
from zsl.task.job_context import JobContext, StatusCodeResponder, WebJobContext, add_responder
from zsl.task.task_decorator import ForbiddenError, json_output
from zsl.utils.documentation import documentation_link
from zsl.utils.http import get_http_status_code_value


class ErrorResponse(AppModel):
    def __init__(self, code, message):
        super(ErrorResponse, self).__init__({})
        self.code = code
        self.message = message


def register(e):
    # type: (ErrorHandler|ErrorProcessor)->None
    if isinstance(e, ErrorHandler):
        _error_handlers.append(e)
    if isinstance(e, ErrorProcessor):
        _error_processors.append(e)


class DefaultErrorHandler(ErrorHandler):
    ERROR_CODE = "UNKNOWN_ERROR"
    ERROR_MESSAGE = "An error occurred!"

    def can_handle(self, e):
        return True

    @json_output
    def handle(self, ex):
        logger = logging.getLogger(__name__)
        logger.error(str(ex) + "\n" + traceback.format_exc())
        logger.error("Request:\n{0}\n{1}\n".format(request.headers,
                                                   request.data))
        link = documentation_link('error_handling')
        logger.info("Provide your own error handler so that "
                    "a better error is produced, check {0}.".format(link))

        add_responder(StatusCodeResponder(http.client.INTERNAL_SERVER_ERROR))
        return ErrorResponse(self.ERROR_CODE, self.ERROR_MESSAGE)


class RoutingErrorHandler(ErrorHandler):
    ERROR_CODE = 'NOT_FOUND'

    def can_handle(self, e):
        return isinstance(e, RoutingError)

    @json_output
    def handle(self, ie):
        logging.error(str(ie) + "\n" + traceback.format_exc())
        add_responder(StatusCodeResponder(get_http_status_code_value(http.client.NOT_FOUND)))
        return ErrorResponse(self.ERROR_CODE, str(ie))


class ModelConversionErrorHandler(ErrorHandler):
    ERROR_CODE = 'INVALID_REQUEST'

    def can_handle(self, e):
        return isinstance(e, ModelConversionError)

    @json_output
    def handle(self, e):
        logging.error(str(e) + "\n" + traceback.format_exc())
        add_responder(StatusCodeResponder(get_http_status_code_value(http.client.UNPROCESSABLE_ENTITY)))
        return ErrorResponse(self.ERROR_CODE, str(e))


class ForbiddenErrorHandler(ErrorHandler):
    ERROR_CODE = 'FORBIDDEN'

    def can_handle(self, e):
        return isinstance(e, ForbiddenError)

    @json_output
    def handle(self, e):
        add_responder(StatusCodeResponder(get_http_status_code_value(http.client.FORBIDDEN)))
        return ErrorResponse(self.ERROR_CODE, str(e))


_DEFAULT_ERROR_HANDLER = DefaultErrorHandler()
_ROUTING_ERROR_HANDLER = RoutingErrorHandler()
_FORBIDDEN_ERROR_HANDLER = ForbiddenErrorHandler()
_MODEL_CONVERSION_ERROR_HANDLER = ModelConversionErrorHandler()
_error_handlers = [_MODEL_CONVERSION_ERROR_HANDLER, _FORBIDDEN_ERROR_HANDLER,
                   _ROUTING_ERROR_HANDLER]  # type: List[ErrorHandler]
_error_processors = []


def error_handler(f):
    """
    Default error handler.
     - On server side error shows a message
       'An error occurred!' and returns 500 status code.
     - Also serves well in the case when the resource/task/method
       is not found - returns 404 status code.
    """

    @wraps(f)
    def error_handling_function(*args, **kwargs):
        @inject(error_config=ErrorConfiguration)
        def get_error_configuration(error_config):
            # type:(ErrorConfiguration)->ErrorConfiguration
            return error_config

        def should_skip_handling():
            use_flask_handler = get_error_configuration().use_flask_handler
            is_web_request = isinstance(JobContext.get_current_context(), WebJobContext)
            return use_flask_handler and is_web_request

        try:
            return f(*args, **kwargs)
        except Exception as ex:
            if should_skip_handling():
                raise

            for ep in _error_processors:
                ep.handle(ex)

            for eh in _error_handlers:
                if eh.can_handle(ex):
                    return eh.handle(ex)

            return _DEFAULT_ERROR_HANDLER.handle(ex)

    return error_handling_function



from __future__ import absolute_import, division, print_function, unicode_literals

from builtins import *
import http.client
from unittest.case import TestCase

from zsl.application.containers.web_container import WebContainer
from zsl.application.modules.web.cors import CORSConfiguration
from zsl.router.task import TaskConfiguration
from zsl.task.task_decorator import json_input, secured_task
from zsl.testing.db import IN_MEMORY_DB_SETTINGS
from zsl.testing.http import HTTPTestCase
from zsl.testing.zsl import ZslTestCase, ZslTestConfiguration
from zsl.utils.security_helper import TOKEN_HASHED, TOKEN_RANDOM, TOKEN_SERVICE_SECURITY_CONFIG, compute_token


class SecuredTask(object):

    @json_input
    @secured_task
    def perform(self, data):
        return "ok"


class SecuredTaskTestCase(ZslTestCase, HTTPTestCase, TestCase):
    CONFIG = IN_MEMORY_DB_SETTINGS.copy()
    CONFIG.update({
        'CORS': CORSConfiguration('origin'),
        'TASKS': TaskConfiguration()
            .create_namespace('test')
                .add_routes({'secured_task': SecuredTask})
            .get_configuration(),
        TOKEN_SERVICE_SECURITY_CONFIG: 'ahoj'
    })

    ZSL_TEST_CONFIGURATION = ZslTestConfiguration(
        app_name='SecuredTaskTestCase',
        container=WebContainer,
        config_object=CONFIG
    )

    def testSecuredTaskWithCorrectSecurityToken(self):
        with self.getHTTPClient() as client:
            random_token = '1'

            response = self.requestTask(client, '/test/secured_task', {
                'security': {
                    TOKEN_RANDOM: random_token,
                    TOKEN_HASHED: compute_token(random_token)
                },
            })

            self.assertHTTPStatus(http.client.OK, response.status_code, 'HTTP status should be 200.')

    def testSecuredTaskWithIncorrectSecurityToken(self):
        with self.getHTTPClient() as client:
            random_token = '1'

            response = self.requestTask(client, '/test/secured_task', {
                'security': {
                    TOKEN_RANDOM: random_token,
                    TOKEN_HASHED: 'incorrect hash'
                },
            })

            self.assertHTTPStatus(http.client.INTERNAL_SERVER_ERROR, response.status_code, 'HTTP status should be 500.')



from __future__ import absolute_import, division, print_function, unicode_literals

from builtins import *
from collections import namedtuple
import logging


class SentryConfiguration(namedtuple('SentryConfiguration', ['dsn', 'environment', 'tags', 'register_logging_handler',
                                                             'sentry_logging_handler_level'])):
    __slots__ = ()

    def __new__(cls, dsn, environment=None, tags=None, register_logging_handler=True,
                sentry_logging_handler_level=logging.ERROR):
        tags = [] if tags is None else tags
        return super(SentryConfiguration, cls).__new__(cls, dsn, environment, tags, register_logging_handler,
                                                       sentry_logging_handler_level)



"""
:mod:`zsl.tasks.asl.schedule_gearman_task`
------------------------------------------
"""
from __future__ import unicode_literals

from builtins import object

from zsl.interface.celery.worker import create_celery_app, zsl_task
from zsl.task.task_data import TaskData
from zsl.task.task_decorator import json_input


def schedule_celery_task(path, data):
    # type: (str, TaskData)->None
    zsl_task.delay({"path": path, "data": data})


class ScheduleCeleryTask(object):
    @json_input
    def perform(self, data):
        # type: (TaskData)->str
        data = data.payload
        schedule_celery_task(data["path"], data["data"])
        return "job submitted"
